-- pointer.e
-- pointer routines by Jeffrey Fielding
-- Please send any bug reports, questions, or comments to JJProg@cyberbury.net
-- version 1.0
-- Types:
-- pointer
--  A pointer (can be NULL) to some data
-- notNullPointer
--  A pointer (can't be NULL) to some data
-- nullPointer
--  A NULL pointer (no data for that pointer)
-- Constants:
-- NULL
--  A generic NULL pointer
-- Routines:
-- notNullPointer p = newPointer()
--  Creates a new pointer
-- newPointerRef(pointer oldPointer)
--  increases the counter value for the pointer (used for garbage collection
--  and recycling)
-- object o = getPointerData(pointer p)
--  returns the data at p (or NULL if p is NULL)
-- setPointerData(notNullPointer p, object value)
--  sets the data at p to value
-- pointer new = clonePointer(pointer p)
--  creates a new pointer, copying the data in p to the new pointer's value
--  so changing the value at new will not change the value at p
-- removePointer(pointer p)
--  removes a reference to the data at p. This is important for garbage
--  collection and recycling. When you call newPointer, you get a pointer
--  with 1 reference. For each new reference to the pointer, you must use
--  newPointerRef, unless you don't use any garbage collecting routines
--  when you call removePointer, it removes one reference to p
--  if the reference count is now zero, calling newPointer or clonePointer
--  may replace the data in it.
--  If there are no references to a pointer, it is considered NULL.
-- collectGarbage()
--  removes as many null pointers as possible without making current pointers
--  invalid
global constant NULL = 0
sequence pointerData
global type pointer(integer n)
    return 1
end type
global type notNullPointer(integer n)
    if pointer(n) then
	return pointerData[n][1] > 0
    end if
    return n > 0 and n <= length(pointerData)
end type
global type nullPointer(integer n)
    if n < 1 or n > length(pointerData) then
	return 1
    end if
    return pointerData[n][1] < 1
end type
pointerData = {}
global function newPointer()
    for i = 1 to length(pointerData) do
	if pointerData[i][1] < 1 then
	    pointerData[i][1] = 1
	    pointerData[i][2] = 0
	    return i
	end if
    end for
    pointerData = append(pointerData,{1,0})
    return length(pointerData)
end function
global procedure newPointerRef(pointer p)
    if not nullPointer(p) then
	pointerData[p][1] = pointerData[p][1] + 1
    end if
end procedure
global function getPointerData(pointer p)
    if nullPointer(p) then
	return NULL
    end if
    return pointerData[p][2]
end function
global procedure setPointerData(notNullPointer p, object v)
    pointerData[p][2] = v
end procedure
global function clonePointer(pointer p)
    notNullPointer np
    if nullPointer(p) then
	return NULL
    else
	np = newPointer()
	setPointerData(np,getPointerData(p))
	return np
    end if
end function
global procedure removePointer(pointer p)
    if notNullPointer(p) then
	pointerData[p][1] = pointerData[p][1] - 1
    end if
end procedure
global procedure collectGarbage()
    integer e
    e = length(pointerData)
    while e > 0 do
	if pointerData[e][1] > 0 then
	    exit
	end if
	e = e - 1
    end while
    if e = 0 then
	pointerData = {}
    else
	pointerData = pointerData[1..e]
    end if
end procedure
